 /*
 * Copyright 1998-2007 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2007 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
#include <linux/config.h>
#endif
#include <linux/ioport.h>
#include <linux/fb.h>
#include <asm/io.h>


#include "share.h"
#include "chip.h"
#include "lcd.h"
#include "tv.h"
#include "ioctl.h"
#include "debug.h"

/* Extern variable */
extern int CRT_ON, TV_ON, DVI_ON, LCD_ON,SAMM_ON;
extern int lcd_panel_id;
extern long tv_out_signal;
extern long tv_system;
extern int DeviceStatus;

/* Extern struct */
extern struct chip_information chip_info;
extern struct tv_setting_information tv_setting_info;

/* Extern function */
extern void write_reg_mask(u8 index, int io_port, u8 data, u8 mask);
extern u8 read_reg(int io_port, u8 index);
extern void write_reg(u8 index, u16 io_port, u8 data);
extern int tv_sense(void);
extern int dvi_sense(void);
extern bool lcd_get_mobile_state(bool* mobile);




/*  
 *   Purpose:
 *       Get device support state.
 *   Parameter:
 *       support_state: a bitwise value used to indicate which device can be supported.
 */
void get_device_support_state(u32* support_state)
{  
    *support_state = CRT_Device;
    if ((chip_info.tv_chip_info.tv_chip_name == VT1622A) ||
        (chip_info.tv_chip_info.tv_chip_name == VT1623) ||
        (chip_info.tv_chip_info.tv_chip_name == VT1625) ||
        (chip_info.tv_chip_info.tv_chip_name == INTEGRATED_TV))
    {
        *support_state |= (TV_Device | HDTV_Device);
    }
    
    if (chip_info.tmds_chip_info.tmds_chip_name == VT1632_TMDS)
    {
        *support_state |= DVI_Device;
    }
    
    if (chip_info.lvds_chip_info.lvds_chip_name == VT1631_LVDS)
    {
        *support_state |= LCD_Device;
    }
}



/* 
 *    Purpose:
 *        Get device connect state.
 *    Parameter:
 *        connect_state: a bitwise value used to indicate which device are connected.
 */
void get_device_connect_state(u32* connect_state)
{   
    bool mobile = FALSE;
    
    *connect_state = CRT_Device; /* default CRT always connect */
    
    if (tv_sense())
    {        
        *connect_state |= (TV_Device | HDTV_Device);
    }
    
    if (dvi_sense())
    {
        *connect_state |= DVI_Device;
    }
    
    lcd_get_mobile_state(&mobile); /* read BIOS */
    if (mobile)
    {
        *connect_state |= LCD_Device;
    }
}



/* 
 *    Purpose:
 *       To know if lcd can support expand function.
 *    Parameter:
 *       xres: horizontal resolution.
 *       yres: vertical resolution.
 *    Return:
 *       A boolean value used to indicate if expand function can be support.
 */
bool lcd_get_support_expand_state(u32 xres, u32 yres)
{
    unsigned int support_state = 0;
    
    switch (lcd_panel_id)
    {
        case LCD_PANEL_ID0_640X480:
            if ((xres < 640) && (yres < 480))
            {
                support_state = TRUE;
            }
            break;
            
        case LCD_PANEL_ID1_800X600:
            if ((xres < 800) && (yres < 600))
            {
                support_state = TRUE;
            }            
            break;
            
        case LCD_PANEL_ID2_1024X768:
            if ((xres < 1024) && (yres < 768))
            {
                support_state = TRUE;
            }
            break;
            
        case LCD_PANEL_ID3_1280X768:
            if ((xres < 1280) && (yres < 768))
            {
                support_state = TRUE;
            }
            break;
            
        case LCD_PANEL_ID4_1280X1024:
            if ((xres < 1280) && (yres < 1024))
            {
                support_state = TRUE;
            }
            break;
            
        case LCD_PANEL_ID5_1400X1050:
            if ((xres < 1400) && (yres < 1050))
            {
                support_state = TRUE;
            }
            break;
            
        case LCD_PANEL_ID6_1600X1200:
            if ((xres < 1600) && (yres < 1200))
            {
                support_state = TRUE;
            }
            break;
            
        case LCD_PANEL_ID7_1366X768:
            if ((xres < 1366) && (yres < 768))
            {
                support_state = TRUE;
            }
            break;
            
        default:
            support_state = FALSE;
            break;
    }

    return support_state;
}


/*=======================================================================================*/
/*                      TV Function Implementation                                       */
/*=======================================================================================*/

/* 
 *    Purpose:
 *        To know which tv standard we can support.
 *    Parameter:
 *        support_state: a bitwise value used to indicate which tv standard we can support.
 */
void tv_get_support_standard(u32* support_state)
{
    switch (chip_info.tv_chip_info.tv_chip_name)
    {
        case VT1622A:
        case VT1623:
            *support_state = TVTYPE_NTSC | TVTYPE_PAL | TVTYPE_480P | TVTYPE_576P;
            break;
        case VT1625:
        case INTEGRATED_TV:
            *support_state = TVTYPE_NTSC | TVTYPE_PAL | TVTYPE_480P | TVTYPE_576P | TVTYPE_720P | TVTYPE_1080I;
            break;
    }
}


/* 
 *    Purpose:
 *        To know which tv signal we can support.
 *    Parameter:
 *        support_state: a bitwise value used to indicate which tv signal we can support.
 */
void tv_get_support_signal(u32* support_state)
{
    switch (chip_info.tv_chip_info.tv_chip_name)
    {
            case VT1621:
                *support_state = TV_OUTPUT_COMPOSITE | TV_OUTPUT_SVIDEO;
                break;
            
            case VT1622:
            case VT1622A:
            case VT1623:
            case VT1625:
            case INTEGRATED_TV:
                *support_state = TV_OUTPUT_COMPOSITE | TV_OUTPUT_SVIDEO | TV_OUTPUT_RGB | TV_OUTPUT_YPBPR
                                | TV_OUTPUT_COMPOSITE_SVIDEO | TV_OUTPUT_RGB_COMPOSITE | TV_OUTPUT_YPBPR_COMPOSITE
                                | TV_OUTPUT_RGB_COMPOSITE_SVIDEO| TV_OUTPUT_YPBPR_COMPOSITE_SVIDEO;
                break;
            
            default:
                *support_state = 0;
                break;
    }
}



/* 
 *    Purpose:
 *        To know what max degree of tv size is.
 *    Return:
 *        Max value of tv size.
 */
int tv_get_max_size(void)
{
    int max_size;
    switch (chip_info.tv_chip_info.tv_chip_name)
    {
        case VT1622A:
        case VT1623:
        case VT1625:
            max_size = TV_SIZE_LEVEL_MAXIMUM;
            break;
            
        case INTEGRATED_TV:
            max_size = 20;
            break;
            
        default:
            max_size = 0;
            break;
    }
    return max_size;
}



/* 
 *    Purpose:
 *        To know which tv tuning operations we can support.
 *    Parameter:
 *        support_state: a bitwise value used to indicate which tv tuning operations we can support.
 */
void tv_get_support_tuning_state(u32* support_state)
{
    switch (chip_info.tv_chip_info.tv_chip_name)
    {
        case VT1622A:
        case VT1623:
        case VT1625:
        case INTEGRATED_TV:
            *support_state = OP_TV_TUNING_FFILTER | OP_TV_TUNING_ADAPTIVE_FFILTER |
                OP_TV_BRIGHTNESS | OP_TV_CONTRAST | OP_TV_SATURATION;
            if ((tv_out_signal != TV_OUTPUT_NONE) && !(tv_out_signal & (TV_OUTPUT_RGB | TV_OUTPUT_YPBPR)))
            {
                *support_state |= OP_TV_TINT;
            }
            break;
        
        default:
            *support_state = 0;
            break;
    }
}



/* 
 *    Purpose:
 *        To get the max value of tv tuning item.
 *    Parameter:
 *        item: indicate which tuning item we want to get.
 *        value: A pointer to max value of tuning item.
 */
void tv_get_tuning_max_value(u32 item, u32* value)
{
    switch (item)
    {
        case OP_TV_BRIGHTNESS:
            *value = TV_BRIGHTNESS_MAXIMUM;
            break;
        
        case OP_TV_CONTRAST:
            if (chip_info.tv_chip_info.tv_chip_name == INTEGRATED_TV)
            {
                *value = 0x1FF; /* integrated TV use 9 bits to set contrast. */
            }
            else
            {
                *value = TV_CONTRAST_MAXIMUM;
            }            
            break;
        
        case OP_TV_SATURATION:
            if (chip_info.tv_chip_info.tv_chip_name == INTEGRATED_TV)
            {
                *value = 0x3FFFF; /* integrated TV use 18 bits to set saturation. */
            }
            else
            {
                *value = TV_SATURATION_MAXIMUM;
            }
            break;
        
        case OP_TV_TINT:
            *value = TV_TINT_MAXIMUM;
            break;
        
        case OP_TV_TUNING_FFILTER:
            if (chip_info.tv_chip_info.tv_chip_name == VT1625)
            {
                *value = TV_FFILTER_MAXIMUM;
            }
            else if (chip_info.tv_chip_info.tv_chip_name == INTEGRATED_TV)
            {
                *value = 1;
            }
            else
            {
                *value = TV_FFILTER_MAXIMUM- 1;
            }
            break;
        
        case OP_TV_TUNING_ADAPTIVE_FFILTER:
            if (chip_info.tv_chip_info.tv_chip_name == VT1625)
            {
                *value = 63; /* VT1625 only use 6 bits to set threshold */
            }
            else if (chip_info.tv_chip_info.tv_chip_name == INTEGRATED_TV)
            {
                *value = 1023; /* Integrated tv use 10 bit to set threshold. */
            }
            else
            {
                *value = TV_ADAPTIVE_FFILTER_MAXIMUM;
            }
            break;
        
        default:
            *value = 0;
            break;
    }
    
}



/* 
 *    Purpose:
 *        To know which tv setting operations we can support.
 *    Parameter:
 *        support_state: a bitwise value used to indicate which tv setting operations we can support.
 */
void tv_get_support_setting_state(u32* support_state)
{
    *support_state |= OP_TV_SETTING_FFILTER;
    switch (chip_info.tv_chip_info.tv_chip_name){
        case VT1622:
        case VT1622A:
        case VT1623:
        case VT1625:
            *support_state |= OP_TV_SETTING_ADAPTIVE_FFILTER;
            if (tv_system == TVTYPE_NTSC && (tv_out_signal & TV_OUTPUT_COMPOSITE))
            {
                *support_state |= OP_TV_DEDOTCRAWL;
            }
            break;
            
        default:
            *support_state = 0;
        break;
    }
}




/*=======================================================================================*/
/*                      Gamma Function Implementation                                    */
/*=======================================================================================*/

/* 
 *      Purpose:
 *          Set gamma table to adjust screen color.
 *      Parameter:
 *          bpp: bits per pixel, 8 bpp mode can't adjust gamma.
 *          gamma_table: an array that save the color value of RGB.
 */
void set_gamma_table(int bpp, unsigned int* gamma_table)
{
    int i, sr1a;
    int active_device_amount = 0;
    int device_status = DeviceStatus;
    unsigned char color_r[256];
    unsigned char color_g[256];
    unsigned char color_b[256];

    for (i=0; i<sizeof(DeviceStatus)*8; i++)
    {
        if (device_status & 1)
        {
            active_device_amount++;
        }
        device_status >>= 1;
    }
    
    for (i=0; i<256; i++)
    {
        color_r[i] = gamma_table[i] >> 16;
        color_g[i] = (gamma_table[i] >> 8) & 0x00FF;
        color_b[i] = gamma_table[i] & 0x000000FF;
    }

    
    /* 8 bpp mode can't adjust gamma */
    if (!(bpp == 8))
    {
        /* Enable Gamma */
        switch (chip_info.gfx_chip_name)
        {
            case UNICHROME_CLE266:
            case UNICHROME_K400:
                write_reg_mask(SR16, VIASR, 0x80, BIT7);
                break;
                
            case UNICHROME_K800:
            case UNICHROME_P880:
            case UNICHROME_CN900:
            case UNICHROME_CX700:
            case UNICHROME_K8M890:
            case UNICHROME_P4M890:
            case UNICHROME_P4M900:
                write_reg_mask(CR33, VIACR, 0x80, BIT7);
                break;
        }
        sr1a = read_reg(VIASR, SR1A);
        write_reg_mask(SR1A, VIASR, 0x0, BIT0); /* map to IGA1 LUT */
        
        /* Fill IGA1 Gamma Table */
        outb(0, LUT_INDEX_WRITE);
        for (i = 0; i < 256; i++)
        {
            outb(color_r[i], LUT_DATA);
            outb(color_g[i], LUT_DATA);
            outb(color_b[i], LUT_DATA);
        }
        
        /* If adjust Gamma value in SAMM, fill IGA1,IGA2 Gamma table simultanous.*/
        /* Switch to IGA2 Gamma Table */
        if ((active_device_amount>1) &&
            !((chip_info.gfx_chip_name == UNICHROME_CLE266) &&
              (chip_info.gfx_chip_revision <15)))
        {
            write_reg_mask(SR1A, VIASR, 0x01, BIT0);  /* map to IGA2 LUT */    
            write_reg_mask(CR6A, VIACR, 0x02, BIT1);  /* LCD GAMMA Enable */
                    
            /* Fill IGA2 Gamma Table */
            outb(0, LUT_INDEX_WRITE);
            for (i = 0; i < 256; i++)
            {
                outb(color_r[i], LUT_DATA);
                outb(color_g[i], LUT_DATA);
                outb(color_b[i], LUT_DATA);
            }
        }
        write_reg(SR1A, VIASR, sr1a);        
    }
}



/* 
 *      Purpose:
 *          Get current gamma table.
 *      Parameter:
 *          gamma_table: an array which save the color value.
 */
void get_gamma_table(unsigned int* gamma_table)
{
    unsigned char color_r[256];
    unsigned char color_g[256];
    unsigned char color_b[256];
    unsigned char sr1a = 0;
    int i;

    /* Enable Gamma */
        switch (chip_info.gfx_chip_name)
        {
            case UNICHROME_CLE266:
            case UNICHROME_K400:
                write_reg_mask(SR16, VIASR, 0x80, BIT7);
                break;
                
            case UNICHROME_K800:
            case UNICHROME_P880:
            case UNICHROME_CN900:
            case UNICHROME_CX700:
            case UNICHROME_K8M890:
            case UNICHROME_P4M890:
            case UNICHROME_P4M900:
                write_reg_mask(CR33, VIACR, 0x80, BIT7);
                break;
        }
        sr1a = read_reg(VIASR, SR1A);
        write_reg_mask(SR1A, VIASR, 0x0, BIT0);
        
    /* Reading gamma table to get color value */
    outb(0, LUT_INDEX_READ);
    for (i = 0; i < 256; i++)
    {
        color_r[i] = inb(LUT_DATA);
        color_g[i] = inb(LUT_DATA);
        color_b[i] = inb(LUT_DATA);
        gamma_table[i] = ((((u32)color_r[i])<<16) | (((u16)color_g[i])<<8)) | color_b[i];
    }
    write_reg(SR1A, VIASR, sr1a);
}


/* 
 *      Purpose:
 *          To know which device we can support for gamma operation.
 *      Parameter:
 *          bpp: bits per pixel, 8 bpp mode can't adjust gamma on any device.
 *          support_state: indicate which device we can support for gamma operation.
 */
void get_gamma_support_state(int bpp, unsigned int* support_state)
{
    if (bpp == 8)
    {
        *support_state = None_Device;
    }
    else
    {
        *support_state = CRT_Device | TV_Device| DVI_Device| LCD_Device;
    }
}

/*
 *     Purpose: convert a serial number into bitwise value.
 */
int input_parameter_converter(int parameter_value)
{    
    int result;

    switch (parameter_value)
    {                    
        case 1:
            result = 0x01;
            break;

        case 2:
            result = 0x02;
            break;            

        case 3:
            result = 0x04;
            break;
            
        case 4:
            result = 0x08;
            break;
            
        case 5:
            result = 0x10;
            break;
            
        case 6:
            result = 0x20;
            break;

        case 7:
            result = 0x40;
            break;
            
        case 8:
            result = 0x80;
            break;

        case 9:
            result = 0x100;
            break;

        default:
            result = 0x01;
            break;
    }

    return result;
}

